(******************************************************************************)
(** * Basic Lemmas about lists  *)
(******************************************************************************)

Require Import PeanoNat Lia.
Require Import Hahn.

Require Import List.
Import ListNotations.

Set Implicit Arguments.

Lemma nth_error_nil {X : Type} n : nth_error (@nil X) n = None.
Proof.
destruct n; auto.
Qed.

Lemma firstn_in : forall {X : Type} (n : nat) (l : list X) x,
                  In x (firstn n l) -> In x l.
Proof.
induction n; ins; desf.
destruct H; vauto.
right; eauto.
Qed.

Lemma skipn_in : forall {X : Type} (n : nat) (l : list X) x,
                  In x (skipn n l) -> In x l.
Proof.
induction n.
auto.
ins.
simpl.
destruct l.
vauto.
right; eauto.
Qed.

Definition Inl {A : Type} (l : list A) : A -> Prop :=
  fun x => In x l.

Definition Inl_pre (l : list nat) : nat -> Prop :=
  fun x => In x l \/ exists y, In y l /\ x < y.

Lemma app_nth_error A n l l' :
  @nth_error A (l ++ l') n =
  if Compare_dec.le_lt_dec (length l) n then nth_error l' (n - length l) else nth_error l n.
Proof.
  desf; eauto using nth_error_app1, nth_error_app2.
Qed.

Definition nth_error_app := app_nth_error. 

Lemma mk_list_nth_error A i n f :
  @nth_error A (mk_list n f) i = if Compare_dec.le_lt_dec n i then None else Some (f i).
Proof.
  induction n; ins; desf; try rewrite app_nth_error; desf;
  try rewrite mk_list_length in *; desf; try lia.
    apply nth_error_None; ins; lia.
    apply nth_error_None; ins; lia.
    replace i with n by lia.
    replace (n - n) with 0 by lia.
    done.
Qed.

Definition nth_error_mk_list := mk_list_nth_error. 

Definition Inb (a : nat) (l : list nat) :=
  existsb (Nat.eqb a) l.

Lemma Inb_In_true a l : In a l <-> Inb a l = true.
Proof.
unfold Inb.
rewrite existsb_exists.
setoid_rewrite NPeano.Nat.eqb_eq.
split; ins; desf; eauto.
Qed.

Lemma Inb_In_false a l : ~ In a l <-> Inb a l = false.
Proof.
destruct (Inb a l) eqn:A; split; ins; auto.
apply Inb_In_true in A; ins. 
intro B; apply Inb_In_true in B; desf.
Qed.

Lemma nth_error_nth {A : Type} n (l : list A) a d :
 nth_error l n = Some a -> nth n l d = a.
Proof.
generalize dependent l.
induction n.
destruct l; ins; desf.
ins; destruct l; ins; desf.
eauto.
Qed.

Lemma nth_error_nth_d {A : Type} n (l : list A) d :
 nth_error l n = None -> nth n l d = d.
Proof.
generalize dependent l.
induction n.
destruct l; ins; desf.
ins; destruct l; ins; desf.
eauto.
Qed.

Lemma nth_error_nth' {A : Type} n (l : list A) d 
 (LEQ : n < length l) :
 nth_error l n = Some (nth n l d).
Proof.
apply nth_error_Some in LEQ.
destruct (nth_error l n) eqn:E; vauto.
erewrite nth_error_nth; auto.
Qed.

Lemma nth_error_nth_len {A : Type} n (l : list A) d
  (IN: n < length l) : nth_error l n = Some (nth n l d).
Proof.
generalize dependent l.
induction n.
destruct l; ins; desf; lia.
ins; destruct l; ins; desf.
lia.
eapply IHn; lia.
Qed.

Lemma nth_neq {A : Type} n (l : list A) d 
  (NEQ: nth n l d <> d) : n < length l.
Proof.
generalize dependent n.
induction l.
destruct n; ins.
destruct n; ins.
lia.
apply IHl in NEQ.
lia.
Qed.

Lemma skipn_length: forall A:Type, forall l:list A, forall n:nat,
  n <= length l -> length (skipn n l) = length l - n.
Proof.
intros A l.
induction l.
intros. destruct n; simpl in *.
- lia.
- auto.
- intros; simpl in *. destruct n. simpl; auto.
simpl. apply IHl. lia.
Qed.

Lemma skipn_nth: forall A:Type, forall l:list A, forall n m:nat, forall d:A,
  nth m (skipn n l) d = nth (n+m) l d.
Proof.
intros A l.
induction l; intros.
by destruct m; destruct n; simpl; auto.
by destruct n; simpl; auto.
Qed.

Lemma skipn_nth_error: forall A:Type, forall l:list A, forall n m:nat,
  nth_error (skipn n l) m = nth_error l (n + m).
Proof.
intros A l; induction l; intros.
by destruct m; destruct n; simpl; auto.
by destruct n; simpl; auto.
Qed.

Lemma nth_error_in_skipn {A : Type} (l:list A) n a :
  nth_error l n = Some a -> In a (skipn n l).
Proof.
generalize dependent l.
induction n.
simpl.
destruct l; ins; auto.
desf; auto.
simpl; destruct l; ins; auto.
Qed.

Lemma nth_error_in_firstn {A : Type} (l:list A) n a :
  nth_error l n = Some a -> In a (firstn (S n) l).
Proof.
generalize dependent l.
induction n.
simpl.
destruct l; ins; auto.
desf; auto.
simpl; destruct l; ins; auto.
Qed.

Lemma firstn_monotone_S {A : Type} (l:list A) n a :
  In a (firstn n l) -> In a (firstn (S n) l).
Proof.
generalize dependent l.
induction n.
by simpl; ins.
ins.
destruct l; auto.
destruct H; vauto.
specialize (IHn l H).
vauto.
Qed.

Lemma firstn_monotone {A : Type} (l:list A) n1 n2 (LEQ: n1 <= n2) a :
  In a (firstn n1 l) -> In a (firstn n2 l).
Proof.
induction LEQ; auto.
intros; apply firstn_monotone_S; auto.
Qed.

Lemma skipn_monotone_S {A : Type} (l:list A) n a :
  In a (skipn (S n) l) -> In a (skipn n l).
Proof.
generalize dependent l.
induction n.
simpl; destruct l; ins; auto.
ins.
destruct l; ins; auto.
Qed.

Lemma skipn_monotone {A : Type} (l:list A) n1 n2 (LEQ: n2 <= n1) a :
  In a (skipn n1 l) -> In a (skipn n2 l).
Proof.
induction LEQ; auto.
ins; apply IHLEQ, skipn_monotone_S; auto.
Qed.

Lemma nth_error_firstn (A : Type) (l : list A) i n
(LEQ : n < i) :
nth_error (firstn i l) n = nth_error l n.
Proof.
generalize dependent n.
generalize dependent i.
induction l.
by destruct n,i; ins.
destruct n,i; ins; try lia.
eapply IHl. lia.
Qed.

Lemma length_filter {A : Type} (l : list A) (f : A -> bool) :
                    length (filter f l) <= length l.
Proof.
induction l; ins; desf; simpl; lia.
Qed.

Lemma length_filter_in {A : Type} (l : list A) (f : A -> bool) 
                       (E: exists x, In x l /\ f x = false) : 
                       length (filter f l) < length l.
Proof. 
induction l.
by desf.
simpl; desf.
2: forward apply (length_filter l f); lia.
simpl in E; desf.
simpl. 
apply -> Nat.succ_lt_mono. eauto.
Qed.

Lemma nth_error_Some' {A : Type} (l : list A) (n : nat) a: 
 nth_error l n = Some a -> n < length l.
Proof.
 intro; desf.
 apply nth_error_Some.
 intro; desf.
Qed.

Lemma nth_error_nth_cases:
  forall (A : Type) i (l : list A) d,
  nth_error l i =
  (if Compare_dec.le_lt_dec (length l) i then None else Some (nth i l d)).
Proof.
ins; desf.
by apply nth_error_None.
by apply nth_error_nth_len.
Qed.

Lemma nth_error_from_nth:
  forall (A : Type) i (l : list A) a d,
  nth i l d = a -> a <> d ->
  nth_error l i = Some a.
Proof.
ins; desf.
apply nth_error_nth_len.
eby eapply nth_neq.
Qed.


Lemma skipn_nil2 (A : Type) (l : list A) n : skipn n l = [] -> length l <= n \/ l = [].
Proof.
generalize dependent l.
induction n; ins; desf; eauto; ins.
apply IHn in H; desf; ins; left; lia.
Qed.

Lemma firstn_mk_list (A : Type) i N (g : nat -> A) (LEQ: N <= i):
  firstn i (mk_list N g) = mk_list N g.
Proof.
induction N; ins.
apply firstn_nil.
rewrite firstn_app.
rewrite IHN; try lia.
f_equal.
rewrite length_mk_list.
destruct (i - N) eqn:E; try lia.
ins.
f_equal.
apply firstn_nil.
Qed.

Lemma map_mk_list (A B : Type) N (f : A -> B) (g : nat -> A) :
  map f (mk_list N g) = mk_list N (fun n => f (g n)).
Proof.
induction N; ins.
rewrite map_app; ins.
by rewrite IHN.
Qed.

Lemma skipn_mk_list (A : Type) N n (LEQ : n <= N) (f : nat -> A) (d : A) :
  skipn n (mk_list N f) = mk_list (N - n) (fun k => f (n + k)).
Proof.
eapply nth_ext with (d:=d).
rewrite skipn_length.
autorewrite with calc_length.
done.
autorewrite with calc_length.
done.
ins.
rewrite skipn_nth.
rewrite !nth_mk_list; desf; lia.
Qed.

Lemma nth_firstn (A : Type) (l : list A) i n (LEQ : n < i) d :
nth n (firstn i l) d = nth n l d.
Proof.
generalize dependent i. 
generalize dependent l. 
generalize dependent n.
induction n.
- ins. destruct i; ins; try lia. destruct l; ins.
- ins. destruct i; ins; try lia. destruct l; ins.
  apply IHn. lia.
Qed. 

Lemma list_func_ext (A : Type) (l1 l2 : list A) 
  (SAME: forall n, nth_error l1 n = nth_error l2 n) : l1 = l2.
Proof.
generalize dependent l2.
induction l1.
- ins; specialize (SAME 0); ins; desf.
- ins. destruct l2.
  + specialize (SAME 0); ins; desf.
  + rewrite IHl1 with (l2 := l2).
    * specialize (SAME 0); ins; desf.
    * ins. specialize (SAME (S n)); ins.
Qed.

Lemma nth_map2 (A B : Type) (l : list A) (f : A -> B) n a b
 (LEN: n < length l): 
nth n (map f l) a = f (nth n l b).
Proof.
generalize dependent n.
induction l; ins; try lia.
destruct n; ins.
apply IHl.
lia.
Qed.

