(******************************************************************************)
(** * Addition of events to execution graphs  *)
(******************************************************************************)

Require Import Lia.

Require Import Hahn.
Require Import AuxRel.

Require Import Labels Events Execution.

Set Implicit Arguments.

Section FG.

Notation "x .( y )" := (y x) (at level 0, only parsing).

(******************************************************************************)
(** * The initial execution graph  *)
(******************************************************************************)

Definition Ginit := 
  {| acts := map (fun x => InitEvent x) Labels.Loc_list ;
    rf := ∅₂ ;
    co := ∅₂ |}.

Lemma Ginit_acts_set : 
  @acts_set Ginit ≡₁ is_init.
Proof.
unfold acts_set, Ginit, is_init; ins.
unfolder; splits; ins.
apply in_map_iff in H; desf; eauto.
desf; apply in_map_iff.
exists x0; splits; auto.
apply Loc_list_prop.
Qed.

Lemma Ginit_wf : Wf Ginit.
Proof.
split; try basic_solver.
all: try by (rewrite Ginit_acts_set; basic_solver).
ins.
rewrite Ginit_acts_set.
unfolder; ins; desf.
by exfalso; apply NEQ, init_same_loc.
Qed.

Lemma Ginit_complete : complete Ginit.
Proof.
unfold complete; rewrite Ginit_acts_set; unfolder; ins; desf.
destruct x; ins.
Qed.

Lemma Ginit_sb : Ginit.(sb) ≡ ∅₂.
Proof.
split; [|basic_solver].
unfold sb, ext_sb.
rewrite Ginit_acts_set.
unfolder; ins; desf; apply in_mk_list_iff in H; apply in_mk_list_iff in H1; desf.
Qed.

Lemma Ginit_hb : Ginit.(hb) ≡ ∅₂.
Proof.
split; [|basic_solver].
unfold hb.
rewrite Ginit_sb.
unfold Ginit.
rels; apply tc_empty.
Qed.

Lemma Ginit_hb_sc : Ginit.(hb_sc) ≡ ∅₂.
Proof.
split; [|basic_solver].
unfold hb_sc.
rewrite Ginit_sb.
unfold fr; unfold Ginit; simpl.
rels; apply tc_empty.
Qed.

(******************************************************************************)
(** * Adding a write to the CO relation  *)
(******************************************************************************)

Definition add_co (co : relation Event) w_prev e :=
  co ∪ (dom_rel (co^? ⨾ ⦗ eq w_prev ⦘) × eq e)
     ∪ (eq e × codom_rel (⦗ eq w_prev⦘ ⨾ co)).

Lemma add_co_trans co w_prev e (TRANS: transitive co) (IRR: irreflexive co) 
  (NIN1: ~ dom_rel co e) (NIN2: ~ codom_rel co e) (NEQ: e <> w_prev):
  transitive (add_co co w_prev e).
Proof.
unfold add_co.
apply transitiveI; relsf; unionL.
- unionR left -> left. done.
- unfolder in *; ins; exfalso; desf; eauto.
- unionR right; basic_solver.
- unionR left -> right; basic_solver 12.
- unfolder in *; ins; exfalso; desf; eauto.
- unfolder in *; ins; exfalso; desf; eauto.
- unfolder in *; ins; exfalso; desf; eauto.
- unionR left -> left; basic_solver.
- unfolder in *; ins; exfalso; desf; eauto.
Qed.

Lemma add_co_total A co w_prev e (TOT: is_total A co) (IN: A w_prev):
  is_total (A ∪₁ eq e) (add_co co w_prev e).
Proof.
unfold add_co.
unfolder; ins; desf.
- apply TOT in NEQ; eauto. tauto.
- assert (co w_prev b \/ co^? b w_prev).
  { destruct (classic (b = w_prev)).
    - basic_solver.
    - apply TOT in H; try done. unfolder; tauto. }
  unfolder in H; basic_solver 12.
- assert (co w_prev a \/ co^? a w_prev).
  { destruct (classic (a = w_prev)).
    - basic_solver.
    - apply TOT in H; try done. unfolder; tauto. }
  unfolder in H; basic_solver 12.
Qed.

Lemma add_co_irr co w_prev e (IRR: irreflexive co)
  (NIN1: ~ dom_rel co e) (NIN2: ~ codom_rel co e) (NEQ: e <> w_prev):
  irreflexive (add_co co w_prev e).
Proof.
unfold add_co.
unfolder; ins; desf.
eapply IRR; eauto.
unfolder in *; basic_solver.
unfolder in *; basic_solver.
Qed.

(******************************************************************************)
(** * Adding an event to an execution graph  *)
(******************************************************************************)

Definition add G G' t n l w :=
 ⟪ IND: forall e, acts_set G e -> tid e = t -> index e < n ⟫ /\
 ⟪ INw: acts_set G w ⟫ /\
 ⟪ LABw: is_w w ⟫ /\
 ⟪ LOCw: loc w = loc_l l ⟫ /\
 ⟪ VALw: is_r_l l -> valw w = valr_l l ⟫ /\
 ⟪ ACTS: acts_set G' ≡₁ acts_set G ∪₁ (eq (ThreadEvent t n l)) ⟫ /\
 ⟪ RF1: is_r_l l -> G'.(rf) ≡ G.(rf) ∪ singl_rel w (ThreadEvent t n l) ⟫ /\
 ⟪ RF2: ~ is_r_l l -> G'.(rf) ≡ G.(rf) ⟫ /\
 ⟪ CO1: is_w_l l -> G'.(co) ≡ add_co G.(co) w (ThreadEvent t n l) ⟫ /\ 
 ⟪ CO2: ~ is_w_l l -> G'.(co) ≡ G.(co) ⟫.

Lemma add_read_rf G G' t n l w (READ: is_r_l l) (ADD: add G G' t n l w) : 
  G'.(rf) ≡ G.(rf) ∪ singl_rel w (ThreadEvent t n l).
Proof.
cdes ADD; eauto.
Qed.

Lemma add_non_read_rf G G' t n l w (NREAD: ~ is_r_l l) (ADD: add G G' t n l w) : 
  G'.(rf) ≡ G.(rf).
Proof.
cdes ADD; eauto.
Qed.

Lemma add_rf_in G G' t n l w (ADD: add G G' t n l w) : 
  G'.(rf) ⊆ G.(rf) ∪ singl_rel w (ThreadEvent t n l).
Proof.
destruct (is_r_l l) eqn:A.
rewrite add_read_rf; eauto; done.
rewrite add_non_read_rf; eauto; basic_solver.
Qed.

Lemma add_write_co G G' t n l w (WRITE: is_w_l l) (ADD: add G G' t n l w) : 
  G'.(co) ≡ add_co G.(co) w (ThreadEvent t n l).
Proof.
cdes ADD; eauto.
Qed.

Lemma add_non_write_co G G' t n l w (NREAD: ~ is_w_l l) (ADD: add G G' t n l w) : 
  G'.(co) ≡ G.(co).
Proof.
cdes ADD; eauto.
Qed.

Lemma add_co_in G G' t n l w (ADD: add G G' t n l w) : 
  G'.(co) ⊆ add_co G.(co) w (ThreadEvent t n l).
Proof.
destruct (is_w_l l) eqn:A.
rewrite add_write_co; eauto; done.
rewrite add_non_write_co; eauto; unfold add_co; basic_solver.
Qed.

Lemma add_fresh G G' t n l w (ADD: add G G' t n l w) : 
  ~ acts_set G (ThreadEvent t n l).
Proof.
red in ADD; desc.
intro A.
specialize (IND (ThreadEvent t n l) A).
assert (n<n) by (apply IND; done).
lia.
Qed.

Lemma add_acts_set G G' t n l w (ADD: add G G' t n l w) : 
  acts_set G' ≡₁ acts_set G ∪₁ (eq (ThreadEvent t n l)).
Proof.
by red in ADD; desc.
Qed.

Lemma add_acts_set_tid G G' t n l w (ADD: add G G' t n l w) : 
  acts_set G' ∩₁ (is_init ∪₁ is_tid t) ≡₁ 
  acts_set G ∩₁ (is_init ∪₁ is_tid t) ∪₁ (eq (ThreadEvent t n l)).
Proof.
rewrite add_acts_set; try edone.
unfolder; unfold is_tid; split; ins; desf; eauto 10.
Qed.

(******************************************************************************)
(** * Add is "always" possible  *)
(******************************************************************************)

Lemma max_helper_l n m : n <= max n m.
Proof.
destruct (classic (n < m)).
rewrite max_r; lia.
rewrite max_l; lia.
Qed.

Lemma max_helper_r n m : m <= max n m.
Proof.
destruct (classic (n < m)).
rewrite max_r; lia.
rewrite max_l; lia.
Qed.

Lemma exists_next_index G t : exists n, 
  forall e, acts_set G e /\ tid e = t -> index e < n.
Proof.
exists (max_of_list (map index G.(acts)) + 1).
ins; desf.
unfold acts_set in H.
cut (index e <= max_of_list (map index (acts G))).
lia.
induction (G.(acts)).
by desf.
destruct H; try subst.
- unfold map, max_of_list; ins.
apply max_helper_l.
- 
rewrite IHl; [|done].
rewrite map_cons.
unfold max_of_list.
apply max_helper_r.
Qed.

Lemma exists_add G t n l w
 (WF: Wf G)
 (IND: forall e, acts_set G e /\ tid e = t -> index e < n)
 (INw: acts_set G w)
 (LAB: is_w w)
 (LOC: loc w = loc_l l)
 (VAL: is_r_l l -> valw w = valr_l l): 
 exists G', add G G' t n l w.
Proof.
destruct l.
- exists {| acts:=  (ThreadEvent t n ((Aload x v))) :: G.(acts);
           rf := G.(rf) ∪ singl_rel w (ThreadEvent t n (Aload x v));
           co := G.(co) |}.
red; splits; eauto; try type_solver.
unfold acts_set; basic_solver.
- exists {| acts:=  (ThreadEvent t n ((Astore x v))) :: G.(acts);
           rf := G.(rf) ;
           co := add_co G.(co) w (ThreadEvent t n (Astore x v)) |}.
red; splits; eauto; try type_solver.
unfold acts_set; basic_solver.
- exists {| acts:=  (ThreadEvent t n ((Armw x vr vw))) :: G.(acts);
           rf := G.(rf) ∪ singl_rel w (ThreadEvent t n (Armw x vr vw));
           co := add_co G.(co) w (ThreadEvent t n (Armw x vr vw)) |}.
red; splits; eauto; try type_solver.
unfold acts_set; basic_solver.
Qed.

Lemma exists_add' G t l w
 (WF: Wf G)
 (INw: acts_set G w)
 (LAB: is_w w)
 (LOC: loc w = loc_l l)
 (VAL: is_r_l l -> valw w = valr_l l): 
 exists n G', add G G' t n l w.
Proof.
generalize (exists_next_index G t); ins; desc.
exists n.
apply exists_add; eauto.
Qed.

(******************************************************************************)
(** * Add increases the graph (monotone)  *)
(******************************************************************************)

Lemma add_acts_monotone G G' t n l w (ADD: add G G' t n l w) : 
  acts_set G ⊆₁ acts_set G'.
Proof.
rewrite (add_acts_set ADD); try edone.
basic_solver.
Qed.

Lemma add_rf_monotone G G' t n l w (ADD: add G G' t n l w) : 
  rf G ⊆ rf G'.
Proof.
cdes ADD.
destruct (is_r_l l).
rewrite RF1; basic_solver.
rewrite RF2; basic_solver.
Qed.

Lemma add_co_monotone G G' t n l w (ADD: add G G' t n l w) : 
  co G ⊆ co G'.
Proof.
cdes ADD.
destruct (is_w_l l).
rewrite CO1; unfold add_co; basic_solver.
rewrite CO2; basic_solver.
Qed.

Lemma add_fr_monotone G G' t n l w (WF: Wf G) (ADD: add G G' t n l w) : 
  fr G ⊆ fr G'.
Proof.
unfold fr.
unfolder; ins; desf; splits; eauto.
eexists.
split.
eby eapply add_rf_monotone.
eby eapply add_co_monotone.
Qed.

Lemma add_sb_monotone G G' t n l w (ADD: add G G' t n l w) : 
  sb G ⊆ sb G'.
Proof.
unfold sb.
rewrite add_acts_monotone; edone.
Qed.

Lemma add_hb_monotone G G' t n l w (ADD: add G G' t n l w) : 
  hb G ⊆ hb G'.
Proof.
unfold hb.
rewrite add_sb_monotone, add_rf_monotone; edone.
Qed.

Lemma add_hb_sc_monotone G G' t n l w (WF: Wf G) (ADD: add G G' t n l w) : 
  hb_sc G ⊆ hb_sc G'.
Proof.
unfold hb_sc.
rewrite add_sb_monotone, add_rf_monotone, add_co_monotone, add_fr_monotone; edone.
Qed.

(******************************************************************************)
(** * Effect of add on derived relations  *)
(******************************************************************************)

Lemma add_sb G G' t n l w (WF: Wf G) (ADD: add G G' t n l w) : 
  sb G' ≡ sb G ∪ ((is_init ∪₁ G.(acts_set) ∩₁ is_tid t) × eq (ThreadEvent t n l)).
Proof.
unfold sb; rewrite add_acts_set; [|edone].
split; unfolder; ins; desf; eauto.
+ 
red in ADD; desc.
exfalso.
unfold ext_sb in H0.
destruct y; desf.
apply IND in H1; simpls; lia.
+ unfold ext_sb in H0.
destruct x; desf; unfold is_tid; eauto.
+ exfalso. eapply ext_sb_irr; edone.
+ destruct x; desf; unfold ext_sb; eauto.
apply (wf_initE WF) in H; basic_solver.
+ splits; eauto. unfold ext_sb. red in ADD; desc.
destruct x; desf; splits; unfold is_tid in *; eauto.
apply IND in H. unfold Events.index in *; desf.
vauto.
Qed.

Lemma add_read_hb G G' t n l w (WF: Wf G) (READ: is_r_l l)
  (ADD: add G G' t n l w) : 
  hb G' ≡ hb G ∪ 
  (dom_rel ((hb G)^? ⨾ ⦗ is_init ∪₁ G.(acts_set) ∩₁ is_tid t ∪₁ eq w ⦘) × (eq (ThreadEvent t n l))).
Proof.
unfold hb.
rewrite add_sb; try edone.
rewrite add_read_rf; try edone.
arewrite (sb G
 ∪ (is_init
    ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l)
 ∪ (rf G ∪ singl_rel w (ThreadEvent t n l)) ≡
(sb G ∪ rf G) ∪
 ((is_init
    ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l)
 ∪ singl_rel w (ThreadEvent t n l))).
basic_solver 12.
rels.
rewrite ct_union_max; [| |basic_solver].
- clear; basic_solver 10.
- rewrite (dom_l (@wf_sbE G)), (dom_l (wf_rfE WF)).
apply dom_max_elt; unfolder; intro; desf; eapply add_fresh; eauto.
Qed.

Lemma add_read_hb_alt_init G G' t n l w (WF: Wf G) (READ: is_r_l l)
  (ADD: add G G' t n l w) : 
  hb G' ≡ hb G ∪ 
  ((is_init ∪₁ dom_rel ((hb G)^? ⨾ ⦗ G.(acts_set) ∩₁ is_tid t ∪₁ eq w ⦘)) 
  × (eq (ThreadEvent t n l))).
Proof.
rewrite add_read_hb; try edone.
rewrite !id_union.
relsf.
arewrite ((hb G)^? ⨾ ⦗is_init⦘ ≡ ⦗is_init⦘).
by rewrite (no_hb_to_init WF); basic_solver.
basic_solver 12.
Qed.

Lemma add_non_read_hb G G' t n l w (WF: Wf G) (NREAD: ~ is_r_l l)
  (ADD: add G G' t n l w) : 
  hb G' ≡ hb G ∪ 
  (dom_rel ((hb G)^? ⨾ ⦗ is_init ∪₁ G.(acts_set) ∩₁ is_tid t ⦘) × (eq (ThreadEvent t n l))).
Proof.
unfold hb.
rewrite add_sb; try edone.
rewrite add_non_read_rf; try edone.
arewrite (sb G
 ∪ (is_init
    ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l)
 ∪ rf G ≡
(sb G ∪ rf G) ∪
 ((is_init
    ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l))).
basic_solver 12.
rels.
rewrite ct_union_max; [| |basic_solver].
- clear; basic_solver 10.
- rewrite (dom_l (@wf_sbE G)), (dom_l (wf_rfE WF)).
apply dom_max_elt; unfolder; intro; desf; eapply add_fresh; eauto.
Qed.

Lemma add_non_read_hb_alt_init G G' t n l w (WF: Wf G) (NREAD: ~ is_r_l l)
  (ADD: add G G' t n l w) : 
  hb G' ≡ hb G ∪ 
  ((is_init ∪₁ dom_rel ((hb G)^? ⨾ ⦗ G.(acts_set) ∩₁ is_tid t⦘)) 
  × (eq (ThreadEvent t n l))).
Proof.
rewrite add_non_read_hb; try edone.
rewrite !id_union.
relsf.
arewrite ((hb G)^? ⨾ ⦗is_init⦘ ≡ ⦗is_init⦘).
by rewrite (no_hb_to_init WF); basic_solver.
basic_solver 12.
Qed.

Lemma add_hb G G' t n l w (WF: Wf G)
  (ADD: add G G' t n l w) : 
  hb G' ≡ hb G ∪ 
  ((is_init ∪₁ dom_rel ((hb G)^? ⨾ ⦗G.(acts_set) ∩₁ is_tid t⦘)) × (eq (ThreadEvent t n l))) ∪
  ((dom_rel ((hb G)^? ⨾ ⦗ eq w ⦘) ∩₁ (fun _ => is_r_l l)) × (eq (ThreadEvent t n l))).
Proof.
destruct (is_r_l l) eqn:RR.
- rewrite add_read_hb; try edone.
  rewrite (no_hb_to_init WF).
  basic_solver 21.
- rewrite add_non_read_hb; try edone.
  rewrite (no_hb_to_init WF).
  basic_solver 21.
  intro; desf.
Qed.

Lemma add_non_write_co_hb G G' t n l w (WF: Wf G)
  (NWRITE: ~ is_w_l l) (ADD: add G G' t n l w) : 
  co G' ;; (hb G')^? ≡ 
  co G ⨾ (hb G)^? ∪ 
  dom_rel (co G ;; (hb G)^? ⨾ ⦗is_tid t ∪₁ eq w⦘)
    × eq (ThreadEvent t n l).
Proof.
rewrite add_non_write_co; try edone.
rewrite add_read_hb; try edone.
2: type_solver.
split.
- rewrite (wf_co_init WF).
  rewrite (no_hb_to_init WF).
  basic_solver 21.
- rewrite (dom_r (wf_coE WF)) at 2.
  rewrite (dom_r (wf_hbE WF)) at 2.
  basic_solver 21.
Qed.

Lemma add_non_write_dom_co_hb G G' t n l w (WF: Wf G)
  (NWRITE: ~ is_w_l l) (ADD: add G G' t n l w) p : 
  dom_rel (co G' ⨾ (hb G')^? ⨾ ⦗is_tid p⦘) ≡₁
  dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid p⦘)
  ∪₁ dom_rel (co G ;; (hb G)^? ⨾ ⦗eq w⦘) ∩₁ (fun _ => t = p).
Proof.
rewrite <- seqA.
rewrite add_non_write_co_hb; try edone.
unfold is_tid; basic_solver 21.
Qed.

Lemma add_hb_in G G' t n l w (WF: Wf G) 
  (ADD: add G G' t n l w) : 
  hb G' ⊆ hb G ∪ 
  (dom_rel ((hb G)^? ⨾ ⦗ is_init ∪₁ G.(acts_set) ∩₁ is_tid t ∪₁ eq w ⦘) × (eq (ThreadEvent t n l))).
Proof.
destruct (is_r_l l) eqn:A.
rewrite add_read_hb; eauto; done.
rewrite add_non_read_hb; basic_solver 12.
Qed.

Lemma add_read_rt_transp_rf G G' t n l w (WF: Wf G) (READ: is_r_l l)
  (ADD: add G G' t n l w) : 
  (G'.(rf)⁻¹)＊ ≡ (G.(rf)^{-1})＊ ∪ 
    ((eq (ThreadEvent t n l)) × dom_rel ((rf G)＊ ⨾ ⦗ eq w ⦘)).
Proof.
rewrite (add_read_rf READ ADD), <- !transp_rt.
rewrite rtE.
rewrite ct_union_max; [| |basic_solver].
2: by rewrite (dom_l (wf_rfE WF)); apply dom_max_elt; 
     unfolder; intro; desf; eapply add_fresh; eauto.
rewrite !rtE; basic_solver 12.
Qed.

(******************************************************************************)
(** * Properties preserved by add  *)
(******************************************************************************)

Lemma add_preserves_no_co_to_init G G' t n l w (CO: G.(co) ≡ G.(co) ⨾  ⦗fun x => ~ is_init x⦘) 
  (ADD: add G G' t n l w) : G'.(co) ≡ G'.(co) ⨾  ⦗fun x => ~ is_init x⦘.
Proof.
split; [|basic_solver].
destruct (is_w_l l) eqn:A.
- rewrite add_write_co; try edone.
  unfold add_co.
  unionL.
  + rewrite CO at 1; basic_solver.
  + basic_solver 12.
  + rewrite CO at 1; basic_solver 21.
- rewrite add_non_write_co; try edone.
  rewrite CO at 1; basic_solver.
  type_solver.
Qed.

Lemma add_preserves_wf G G' t n l w (WF: Wf G) 
  (ADD: add G G' t n l w) : Wf G'.
Proof.
cdes ADD; desc.
split; ins.
- rewrite add_acts_set; try edone.
rewrite !id_union; relsf; unionL.
* rewrite (wf_index WF); basic_solver.
* unfolder; ins; desf; exfalso.
unfold same_index, same_tid in *.
simpl in *; subst; specialize (IND y H1 eq_refl); lia.
* unfolder; ins; desf; exfalso.
unfold same_index, same_tid in *.
simpl in *; subst; specialize (IND x H eq_refl); lia.
* basic_solver.
- apply dom_helper_3; destruct (is_r_l l).
* rewrite RF1, (wf_rfE WF); try done.
  rewrite ACTS; basic_solver.
* rewrite RF2, (wf_rfE WF); try type_solver.
  rewrite ACTS; basic_solver.
- apply dom_helper_3; destruct l.
* rewrite RF1, (wf_rfD WF); type_solver.
* rewrite RF2, (wf_rfD WF); type_solver.
* rewrite RF1, (wf_rfD WF); type_solver.
- destruct l.
* rewrite RF1, (wf_rfl WF); type_solver.
* rewrite RF2, (wf_rfl WF); type_solver.
* rewrite RF1, (wf_rfl WF); type_solver.
- destruct l.
* apply RF1 in RF; [|type_solver]. unfolder in RF; desf.
  apply WF; basic_solver.
  basic_solver.
* apply RF2 in RF; [|type_solver].
  apply WF; basic_solver.
* apply RF1 in RF; [|type_solver]. unfolder in RF; desf.
  apply WF; basic_solver.
  basic_solver.
- destruct l.
* rewrite RF1; try done; rewrite transp_union.
apply functional_union; [apply WF| basic_solver|].
unfolder; ins; desf.
eapply add_fresh; try edone.
by apply (dom_r (wf_rfE WF)) in H; unfolder in H; desc.
* rewrite RF2; try done; apply WF.
* rewrite RF1; try done; rewrite transp_union.
apply functional_union; [apply WF| basic_solver|].
unfolder; ins; desf.
eapply add_fresh; try edone.
by apply (dom_r (wf_rfE WF)) in H; unfolder in H; desc.
- apply dom_helper_3; rewrite add_acts_set; [|try edone]; destruct (is_w_l l).
+ rewrite CO1; try done; unfold add_co; rewrite (wf_coE WF); basic_solver.
+ rewrite CO2; try done; rewrite (wf_coE WF); basic_solver.
- apply dom_helper_3; destruct l.
+ rewrite CO2; try done; rewrite (wf_coD WF); basic_solver.
+ rewrite CO1; try done; unfold add_co; rewrite (wf_coD WF); basic_solver.
+ rewrite CO1; try done; unfold add_co; rewrite (wf_coD WF); basic_solver.
- destruct l.
+ rewrite CO2; try done; apply WF.
+ rewrite CO1; try done; unfold add_co; rewrite (wf_col WF); basic_solver.
+ rewrite CO1; try done; unfold add_co; rewrite (wf_col WF); basic_solver.
- destruct (is_w_l l).
+ rewrite CO1; try done. apply add_co_trans; try apply WF.
* intro. eapply add_fresh; try edone. unfolder in H; desc.
  apply (wf_coE WF) in H; unfolder in H; desf.
* intro. eapply add_fresh; try edone. unfolder in H; desc.
  apply (wf_coE WF) in H; unfolder in H; desf.
* intro; subst; eapply add_fresh; edone.
+ rewrite CO2; try done. apply WF.
- destruct (is_w_l l) eqn:W.
+ rewrite add_acts_set; try edone.
rewrite CO1; try done.
* destruct (classic (loc_l l = x)).
-- arewrite ((acts_set G ∪₁ eq (ThreadEvent t n l)) ∩₁ (fun a : Event => is_w a)
   ∩₁ (fun a : Event => loc a = x) ⊆₁ (acts_set G ∩₁ (fun a : Event => is_w a)
   ∩₁ (fun a : Event => loc a = x) ∪₁ eq (ThreadEvent t n l))).
by type_solver.
 apply add_co_total. apply WF. basic_solver.
-- unfold add_co. unfolder; ins; desf.
eapply (wf_co_total WF) in NEQ. tauto. basic_solver. basic_solver.
+ rewrite add_acts_set; try edone.
rewrite CO2; try done. 
unfolder; ins; desf.
eapply WF; basic_solver.
- destruct (is_w_l l).
+ rewrite CO1; try done. apply add_co_irr; try apply WF.
* intro. eapply add_fresh; try edone. unfolder in H; desc.
  apply (wf_coE WF) in H; unfolder in H; desf.
* intro. eapply add_fresh; try edone. unfolder in H; desc.
  apply (wf_coE WF) in H; unfolder in H; desf.
* intro; subst; eapply add_fresh; edone.
+ rewrite CO2; try done. apply WF.
- unfolder; ins; eapply add_acts_set; try edone; left; apply WF; done.
- eapply add_preserves_no_co_to_init; try edone. apply WF.
Qed.

Lemma add_preserves_complete G G' t n l w (COMP: complete G) 
  (ADD: add G G' t n l w) : complete G'.
Proof.
cdes ADD; desc.
unfold complete; ins.
rewrite add_acts_set; eauto.
relsf; splits.
- red in COMP; rewrite COMP.
  rewrite add_rf_monotone; eauto.
- destruct (is_r_l l) eqn:A.
  + rewrite RF1; eauto.
    basic_solver.
  + rewrite RF2; eauto.
    type_solver.
Qed.

Lemma add_preserves_rf_to_rmw_in_co G G' t n l w 
  (RF: G.(rf) ⨾ ⦗is_rmw⦘ ⊆ G.(co)) 
  (ADD: add G G' t n l w) : G'.(rf) ⨾ ⦗is_rmw⦘ ⊆ G'.(co).
Proof.
destruct l.
- rewrite add_read_rf; try edone; try type_solver.
  rewrite add_non_write_co; try edone; try type_solver.
  relsf; rewrite RF.
  type_solver 21.
- rewrite add_non_read_rf; try edone; try type_solver.
  rewrite add_write_co; try edone; try type_solver.
  relsf.
- rewrite add_read_rf; try edone; try type_solver.
  rewrite add_write_co; try edone; try type_solver.
  relsf; rewrite RF.
  unfold add_co; basic_solver 21.
Qed.

End FG.
