(******************************************************************************)
(** * SRAG memory subsystem   *)
(******************************************************************************)

Require Import PeanoNat Lt Gt.
Require Import Lia.

Require Import Hahn.
Require Import AuxRel.

Require Import Labels Events Execution FG Wmax Add_max.

Set Implicit Arguments.

Notation "'R'" := is_r.
Notation "'W'" := is_w.
Notation "'RMW'" := is_rmw.

Section SRAG.

Definition SRAGinit G := G = Ginit.

Definition SRAG_step G G' o :=
  exists t l n w,
  ⟪ O: o = Some (t,l) ⟫ /\
  ⟪ WF_G: Wf G ⟫ /\
  ⟪ MAX: is_w_l l -> w_max G (loc_l l) w ⟫ /\
  ⟪ ADD: add G G' t n l w ⟫ /\
  ⟪ CO: ~ is_w_l l -> ~ dom_rel (G.(co) ⨾ (hb G)^? ⨾ ⦗ is_tid t ⦘) w ⟫.

(******************************************************************************)
(** * SRAG preserves properties preserved by adding events    *)
(******************************************************************************)

Lemma SRAG_step_preserves_wf G G' o
  (STEP: SRAG_step G G' o): Wf G'.
Proof.
red in STEP; desc.
eapply add_preserves_wf; edone.
Qed.

Lemma SRAG_step_preserves_complete G G' o (COMP: complete G) 
  (STEP: SRAG_step G G' o): complete G'.
Proof.
red in STEP; desc.
eapply add_preserves_complete; edone.
Qed.

Lemma SRAG_step_preserves_no_co_to_init G G' o
  (CO: G.(co) ≡ G.(co) ⨾  ⦗fun x => ~ is_init x⦘) 
  (STEP: SRAG_step G G' o): G'.(co) ≡ G'.(co) ⨾  ⦗fun x => ~ is_init x⦘.
Proof.
red in STEP; desc.
eapply add_preserves_no_co_to_init; edone.
Qed.

Lemma SRAG_step_preserves_rf_to_rmw_in_co G G' o
  (RF: G.(rf) ⨾ ⦗is_rmw⦘ ⊆ G.(co)) 
  (STEP: SRAG_step G G' o): G'.(rf) ⨾ ⦗is_rmw⦘ ⊆ G'.(co).
Proof.
red in STEP; desc.
eapply add_preserves_rf_to_rmw_in_co; edone.
Qed.

End SRAG.

(******************************************************************************)
(** ** consistency  *)
(******************************************************************************)

Section SRA.

Variable G : execution.

Notation "'E'" := (acts_set G).
Notation "'sb'" := (sb G).
Notation "'rf'" := G.(rf).
Notation "'co'" := G.(co).

Notation "'fr'" := (fr G).
Notation "'coe'" := (coe G).
Notation "'coi'" := (coi G).
Notation "'rfi'" := (rfi G).
Notation "'rfe'" := (rfe G).
Notation "'fre'" := (fre G).
Notation "'fri'" := (fri G).

Notation "'hb'" := (hb G).

Implicit Type WF : Wf G.
Implicit Type COMP : complete G.

Definition sra_consistent :=
  ⟪ IRR : acyclic (hb ∪ co) ⟫ /\
  ⟪ RCOH: irreflexive (hb ⨾ fr) ⟫ /\
  ⟪ ATOM : irreflexive (fr ⨾ co) ⟫.

Lemma WCOH (CONS: sra_consistent) :
  irreflexive (hb ⨾ co).
Proof.
cdes CONS.
unfolder; ins; desf.
eapply IRR.
apply ct_begin.
exists z.
splits; vauto.
Qed.

Lemma immediate_co_helper (WF: Wf G) (COMP: complete G)
  (CONS: sra_consistent): 
  rf ⨾ ⦗RMW⦘ ≡ (immediate co) ⨾ ⦗RMW⦘.
Proof.
cdes CONS.
rewrite immediateE.
rewrite (wf_rfE WF).
rewrite (wf_rfD WF).
rewrite (wf_coE WF) at 1.
rewrite (wf_coD WF) at 1.
unfolder; splits; ins; desf; splits; eauto; try type_solver.
- eapply tot_ex; try eapply WF.
unfolder; splits; eauto; type_solver.
unfolder; splits; eauto; eapply WF; eauto.
intro; eapply (WCOH CONS); exists y; splits; vauto.
intro; subst; eapply IRR; unfold "hb"; left; left; vauto.
- intro; desf.
assert (y <> z) by (intro; subst; eapply WF; eauto).
 eapply (ATOM y). unfold Execution.fr.
unfolder; ins; desf; eauto 21.
exists z; split; eauto; splits; eauto.
intro; subst; desf.
- assert (CR: codom_rel rf y).
eapply COMP; unfolder; splits; try type_solver.
unfolder in CR; desf.
destruct (classic (x=x0)); desf.
exfalso.
eapply (wf_co_total WF) in H6.
+ desf.
*
eapply H1; eexists; splits; eauto.
eapply tot_ex; try eapply WF.
basic_solver.
unfolder; splits; eauto.
apply (wf_rfE WF) in CR; unfolder in CR; desf.
apply (wf_rfD WF) in CR; unfolder in CR; desf.
apply (wf_rfl WF) in CR; unfolder in CR; desf.
intro; eapply WCOH; eexists; splits; vauto.
intro; subst; eapply IRR; left; left; vauto.
* 
 eapply (ATOM y). unfold Execution.fr.
unfolder; ins; desf; eauto 21.
eexists; splits; eauto.
intro; desf; eapply WF; eauto.
+ basic_solver.
+ unfolder; splits; eauto.
apply (wf_rfE WF) in CR; unfolder in CR; desf.
apply (wf_rfD WF) in CR; unfolder in CR; desf.
apply (wf_rfl WF) in CR; unfolder in CR; desf.
apply (wf_col WF) in H2; unfolder in H2; desf.
congruence.
Qed.

Lemma immediate_co_pred_exists (WF: Wf G) (CONS: sra_consistent) 
  e (IN: E e) (WRITE: is_w e) (NINIT: ~ is_init e) :
  exists w, immediate co w e.
Proof.
eapply fsupp_immediate_pred; try eapply WF.
- arewrite (co ⊆ E × E).
eapply dom_helper_3, WF.
eapply fsupp_cross.
by red; unfold acts_set; eauto.
- unfolder; intro; eapply (H (InitEvent (loc e))).
eapply tot_ex; try eapply WF.
+ unfolder; splits; eauto.
+ unfolder; splits; eauto.
eapply WF; vauto.
+ intro.
eapply (WCOH CONS).
eexists; splits; eauto.
econs; left.
unfold Execution.sb, ext_sb; unfolder; splits; eauto.
eapply WF; vauto.
unfold is_init in *; desf.
+ intro; eapply NINIT; unfold is_init; desf.
Qed.


End SRA.

Section SRAG_SRA.

Lemma Ginit_sra_consistent : sra_consistent Ginit.
Proof.
red; unnw; rewrite Ginit_hb. 
unfold Execution.fr, Ginit; ins.
unfold acyclic; relsf; basic_solver.
Qed.

Lemma SRAG_step_preserves_consistency G G' t l (WF: Wf G) 
  (CO: G.(co) ≡ G.(co) ⨾  ⦗fun x => ~ is_init x⦘) 
  (CONS: sra_consistent G)
  (STEP: SRAG_step G G' (Some (t,l))): sra_consistent G'.
Proof.
unfold sra_consistent in *; desf.
unfold Execution.fr.
red in STEP; desc.
unnw.
assert (Ww: W w) by apply ADD.
destruct l eqn:K; desf; ins.
- clear MAX.
  rewrite add_read_hb; try edone; [|type_solver].
  rewrite add_non_write_co; try edone; [|type_solver].
  rewrite add_read_rf; try edone; [|type_solver].
  relsf.
  rewrite !transp_union, !transp_singl_rel, !seq_union_l, !minus_union_l; relsf.
  rewrite !irreflexive_union.
  splits; try edone.

{rewrite unionC, <- unionA, unionC.
apply acyclic_union1.
- rewrite (dom_l (wf_hbE WF)).
apply acyclic_disj.
unfolder; splits; ins; desf; try eby eapply add_fresh. 
- by rewrite unionC; auto.
- arewrite ((co G ∪ hb G)⁺ ⊆  ⦗acts_set G⦘ ;; (co G ∪ hb G)⁺).
{ rewrite (dom_l (wf_hbE WF)) at 1. rewrite (dom_l (wf_coE WF)) at 1.
rewrite <- seq_union_r.
by rewrite inclusion_ct_seq_eqv_l. }
rewrite ct_end, !seqA.
arewrite_false (dom_rel ((hb G)^? ⨾ ⦗is_init ∪₁ acts_set G ∩₁ is_tid t0 ∪₁ eq w⦘) × eq (ThreadEvent t0 n (Aload x v))
   ⨾ ⦗acts_set G⦘).
{ unfolder; ins; desf; eby eapply add_fresh. }
relsf.
red; relsf. }
  all: try rewrite (wf_hbE WF).
  all: try rewrite (wf_coE WF).
  all: try rewrite (wf_rfE WF).
  all: unfolder; ins; desf.
  all: try eby eapply add_fresh.
  by apply CO in H3; unfolder in H3; desc; eauto.
  by apply no_hb_to_init in H8; unfolder in H8; desc; eauto.
  by eapply CO0; basic_solver 12.
  by eapply CO0; basic_solver 12.
  eby eapply WF.
  by eapply WCOH; basic_solver.
- rewrite add_non_read_hb; try edone; [|type_solver].
  rewrite add_max_write_co with (l:= (Astore x v)) (x:=x); try edone.
  rewrite add_non_read_rf with (l:= (Astore x v)); try edone.
  relsf.
split; desf.
{ 
rewrite (dom_l (wf_hbE WF)) at 2. 
remember (dom_rel ((⦗acts_set G⦘ ⨾ hb G)^? ⨾ ⦗is_init ∪₁ acts_set G ∩₁ is_tid t0⦘)) as A.
remember (acts_set G ∩₁ (fun a : Event => W a) ∩₁ is_loc x) as B.
eapply acyclic_mon with (r := hb G ∪ co G ∪ (A ∪₁ B) × eq (ThreadEvent t0 n (Astore x v))).
2: basic_solver.
apply acyclic_union1; auto.
- 
apply acyclic_disj.
subst.
split; unfolder; ins; desf; try eby eapply add_fresh.
- 
rewrite (dom_l (wf_hbE WF)) at 1. 
rewrite (dom_l (wf_coE WF)) at 1.
rewrite <- seq_union_r.
rewrite inclusion_ct_seq_eqv_l.
rewrite ct_end with (r:= ((A ∪₁ B) × eq (ThreadEvent t0 n (Astore x v)))).
apply acyclic_disj.
subst.
split; unfolder; ins; desf; try eby eapply add_fresh.
}

  rewrite minus_union_l, !id_union, !crE.
  rewrite (wf_hbE WF).
  rewrite (wf_coE WF).
  rewrite (wf_rfE WF).
  rewrite (no_hb_to_init WF).
  rewrite CO; desf.
  unfolder; ins; desf; splits; ins; desf; splits.
  all: try eby eapply add_fresh.
  all: try by (eapply WCOH; basic_solver 12).
  all: try by (eapply RCOH; unfold fr; basic_solver 12).
  all: try by (eapply ATOM; unfold fr; basic_solver 12).
  all: try by eapply MAX; ins; edone.

- rewrite add_read_hb; try edone; [|type_solver].
  rewrite add_max_write_co with (l:= (Armw x vr vw)) (x:=x); try edone.
  rewrite add_read_rf; try edone; [|type_solver].
  relsf.
split; desf.
{ 
rewrite (dom_l (wf_hbE WF)) at 2. 
remember (dom_rel ((⦗acts_set G⦘ ⨾ hb G)^? ⨾ ⦗is_init ∪₁ acts_set G ∩₁ is_tid t0 ∪₁ eq w⦘)) as A.
remember (acts_set G ∩₁ (fun a : Event => W a) ∩₁ is_loc x) as B.
eapply acyclic_mon with (r := hb G ∪ co G ∪ (A ∪₁ B) × eq (ThreadEvent t0 n (Armw x vr vw))).
2: basic_solver.
apply acyclic_union1; auto.
- 
apply acyclic_disj.
subst.
split; unfolder; ins; desf; try eby eapply add_fresh.
all: try by eapply add_fresh; try edone; eapply MAX; auto.
- 
rewrite (dom_l (wf_hbE WF)) at 1. 
rewrite (dom_l (wf_coE WF)) at 1.
rewrite <- seq_union_r.
rewrite inclusion_ct_seq_eqv_l.
rewrite ct_end with (r:= ((A ∪₁ B) × eq (ThreadEvent t0 n (Armw x vr vw)))).
apply acyclic_disj.
subst.
split; unfolder; ins; desf; try eby eapply add_fresh.
}
  rewrite !minus_union_l, !id_union, !crE.
  rewrite (wf_hbE WF).
  rewrite (wf_coE WF).
  rewrite (wf_rfE WF).
  rewrite (no_hb_to_init WF).
  rewrite CO; desf.
  unfolder; ins; desf; splits; ins; desf; splits.
  all: try eby (cdes ADD; eapply add_fresh).
  all: try by (eapply WCOH; basic_solver 12).
  all: try by (eapply RCOH; unfold fr; basic_solver 12).
  all: try by (eapply ATOM; unfold fr; basic_solver 12).
  all: try by eapply MAX; try edone. 
Qed.

End SRAG_SRA.

