(******************************************************************************)
(** * SRAL memory subsystem   *)
(******************************************************************************)

Require Import List.
Import ListNotations.

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.

Require Import Basic.
Require Import Labels.
Require Import Subseq.

Set Implicit Arguments.

Section SRALn.

(******************************************************************************)
(** * Read options  *)
(******************************************************************************)

Inductive Opt := 
  | RT (tw : Tid) (v : Val) (rmw_flag : bool).

Definition tid_opt o :=
  match o with
  | RT tw _ _ => tw
  end.

Definition val_opt o :=
  match o with
  | RT _ v _ => v
  end.

Definition rmw_flag_opt o :=
  match o with
  | RT _ _ u => u
  end.

Definition PStore := Labels.Loc -> Opt.

Definition init_ps : PStore := fun x => (RT T0 0 false).

(******************************************************************************)
(** * States  *)
(******************************************************************************)

Definition Potential := list (list PStore).

(* Definition wf_potential (B : Potential) := 
  << NON_EMPTY_POT : B <> [] >> /\
  << NON_EMPTY_OL  : forall L (IN: In L B), L <> [] >> /\
  << RMW_MON : forall L (IN: In L B) n1 n2 (LEQ: n1 < n2) 
               ps1 (IN1: nth_error L n1 = Some ps1)
               ps2 (IN2: nth_error L n2 = Some ps2)
               y (RMW1: rmw_flag_opt (ps1 y) = true), 
               rmw_flag_opt (ps2 y) = true >>.
 *)
Definition State := Tid -> Potential.

Definition wf_state (st : State) := 
  << NON_EMPTY_POT : forall t, st t <> [] >> /\
  << NON_EMPTY_OL  : forall t L (IN: In L (st t)), L <> [] >> /\
  << RMW_MON : forall t L (IN: In L (st t)) n1 n2 (LEQ: n1 < n2) 
               ps1 (IN1: nth_error L n1 = Some ps1)
               ps2 (IN2: nth_error L n2 = Some ps2)
               y (RMW1: rmw_flag_opt (ps1 y) = true), 
               rmw_flag_opt (ps2 y) = true >> /\
  << SAME_END : exists last_val last_tid, 
                forall t L (IN: In L (st t)) y, 
                nth (pred (length L)) L init_ps y = 
                RT (last_tid y) (last_val y)  true >>.

(******************************************************************************)
(** * Steps  *)
(******************************************************************************)

Definition SRAL_read_step_helper (st : State) t x v u :=
  exists tw, forall ol (IN: In ol (st t)) ps 
  (PS: hd_error ol = Some ps), ps x = RT tw v u.

Definition upd_read_ps (ps : PStore) x := 
  upd ps x (RT (tid_opt (ps x)) (val_opt (ps x)) false).

Definition upd_ps (ps : PStore) x t v := upd ps x (RT t v true).

Definition update_ol ol i x t v : list PStore :=
  map (fun ps => upd_read_ps ps x) (firstn i ol) ++
  map (fun ps => upd_ps ps x t v) (skipn i ol).

Definition SRAL_write_step_helper_p B B' Bt t x v p :=
  forall ol' (IN: In ol' B'), 
     exists ol i, 
     i < length ol /\ (p = t -> i = 0) /\ 
     In ol B /\ In (skipn i ol) Bt /\ 
     ol' = update_ol ol i x t v.

Definition SRAL_write_step_helper (st st' : State) t x v :=
 << P : forall p, SRAL_write_step_helper_p (st p) (st' p) (st t) t x v p >> /\
 << NON_EMPTY' : forall p, st' p <> [] >>.

Inductive SRAL_step : State -> State -> option (prod Tid Lab)-> Prop :=
| step_read st t x v u
  (READ: SRAL_read_step_helper st t x v u)
  : SRAL_step st st (Some (t, (Aload x v)))
| step_write st st' t x v
  (WRITE: SRAL_write_step_helper st st' t x v)
   : SRAL_step st st' (Some (t, (Astore x v)))
| step_rmw st st' t x vr vw 
  (READ: SRAL_read_step_helper st t x vr true)
  (WRITE: SRAL_write_step_helper st st' t x vw)
  : SRAL_step st st' (Some (t, (Armw x vr vw)))
| step_dup (st st' : State) 
  (DUP : forall p ol' (IN: In ol' (st' p)), 
     exists ol1 ol2 (NE : length ol2 > 0), 
     In (ol1 ++ ol2) (st p) /\ ol' = ol1 ++ [hd init_ps ol2] ++ ol2)
  (WF': wf_state st')  
  : SRAL_step st st' None
| step_lose (st st' : State) 
  (SUB : forall p ol' (IN: In ol' (st' p)), 
     exists ol, In ol (st p) /\ subseq ol' ol)
  (WF': wf_state st') 
  : SRAL_step st st' None.

(******************************************************************************)
(** * Alternative formulation of write step  *)
(******************************************************************************)

Lemma SRAL_write_step_helper_p_equiv B B' Bt t x v p :
 SRAL_write_step_helper_p B B' Bt t x v p ->
  forall ol' (IN: In ol' B'), 
     exists ol i, 
     i < length ol /\ (p = t -> i = 0) /\ 
     In ol B /\ In (skipn i ol) Bt /\ 
     (forall n (NF: n < i) ps (PS: nth_error ol n = Some ps), 
           nth_error ol' n = Some (upd_read_ps ps x)) /\
     (forall n (NF: n >= i) 
             ps (PS: nth_error ol n = Some ps), 
           nth_error ol' n = Some (upd_ps ps x t v)) /\
     (length ol = length ol').
Proof.
unfold SRAL_write_step_helper_p; ins.
specialize (H _ IN); desc.
  eexists ol,i.
unfold update_ol in *; desf.
splits; auto.
- ins.
rewrite app_nth_error; desf; rewrite length_map in *.
+ rewrite firstn_length_le in *; try lia.
+ rewrite firstn_length_le in *; try lia.
 rewrite nth_error_map.
rewrite nth_error_firstn; try done.
rewrite PS; ins.
- ins.
rewrite app_nth_error; desf; rewrite length_map in *.

+ rewrite firstn_length_le in *; try lia.
 rewrite nth_error_map.

rewrite skipn_nth_error; try done.
replace (i + (n - i)) with n by lia.
rewrite PS; ins.
+ rewrite firstn_length_le in *; try lia.
- rewrite length_app, !length_map.
rewrite firstn_length_le in *; try lia.
rewrite skipn_length; lia.
Qed.




(******************************************************************************)
(** * step preserves well-formedness  *)
(******************************************************************************)

Lemma SRAL_write_step_helper_preserves_wf st st' t x v (WF: wf_state st) 
  (STEP:  SRAL_write_step_helper st st' t x v): wf_state st'.
Proof.
red in STEP; desc.
red in WF; desc.
 
red; splits; eauto.

- ins.
rename L into ol'.
eapply SRAL_write_step_helper_p_equiv in P; try edone.
intro X; desf; ins.
lia.


- ins.
rename L into ol'.
eapply SRAL_write_step_helper_p_equiv in P; try edone; desc.
destruct (classic (n2 < i)).
-- destruct (classic (n1 < i)); try lia.
clear P4.
assert (P31 := P3 _ H0 (nth n1 ol init_ps)).
specialize (P3 _ H (nth n2 ol init_ps)).
rewrite List.nth_error_nth' with (d := init_ps) in P31; try lia.
rewrite List.nth_error_nth' with (d := init_ps) in P3; try lia.
specialize (P31 eq_refl).
specialize (P3 eq_refl).
rewrite IN1 in P31; clear IN1.
rewrite IN2 in P3; clear IN2.
desf.


unfold upd_read_ps, upd in *; desf; ins.
clear P2.
eapply RMW_MON with (n1:=n1) (n2:=n2); try edone;
apply nth_error_nth_len; lia.
-- destruct (classic (n1 < i)); try lia.

++ specialize (P3 _ H0 (nth n1 ol init_ps)).
assert (N2: n2 >= i) by lia.

specialize (P4 _ N2 (nth n2 ol init_ps)).
rewrite List.nth_error_nth' with (d := init_ps) in P3; try lia.
rewrite List.nth_error_nth' with (d := init_ps) in P4; try lia.
specialize (P3 eq_refl).
specialize (P4 eq_refl).
rewrite IN1 in P3.
rewrite IN2 in P4.
desf.


unfold upd_read_ps, upd_ps, upd in *; desf; ins.
clear P2.
eapply RMW_MON with (n1:=n1) (n2:=n2); try edone;
apply nth_error_nth_len; try lia.
eby rewrite P5; eapply nth_error_Some'.
eby rewrite P5; eapply nth_error_Some'.
++ 
clear P3.
assert (N1: n1 >= i) by lia.
assert (N2: n2 >= i) by lia.
assert (P41 := P4 _ N1 (nth n1 ol init_ps)).
specialize (P4 _ N2 (nth n2 ol init_ps)).
rewrite List.nth_error_nth' with (d := init_ps) in P41; try lia.
rewrite List.nth_error_nth' with (d := init_ps) in P4; try lia.
specialize (P41 eq_refl).
specialize (P4 eq_refl).
rewrite IN1 in P41.
rewrite IN2 in P4.
desf.


unfold upd_ps, upd in *; desf; ins.
clear P2.
eapply RMW_MON with (n1:=n1) (n2:=n2); try edone;
apply nth_error_nth_len; try lia.
eby rewrite P5; eapply nth_error_Some'.
eby rewrite P5; eapply nth_error_Some'.
eby rewrite P5; eapply nth_error_Some'.
eby rewrite P5; eapply nth_error_Some'.

-  exists (fun y => ifP (y = x) then v else last_val y).
exists (fun y => ifP (y = x) then t else last_tid y).
ins.


rename t0 into p.
specialize (P p _ IN); desc; subst.
unfold update_ol.
rewrite length_app.
rewrite !length_map.

assert ((length (skipn i ol)) > 0).
{ remember (skipn i ol) as olt.
apply NON_EMPTY_OL in P2.
destruct olt; desf; ins.
lia. }
rewrite app_nth2; rewrite !length_map; try lia.

replace (Init.Nat.pred (length (firstn i ol) + length (skipn i ol)) - length (firstn i ol)) with (Init.Nat.pred (length (skipn i ol))) by lia.
remember (fun ps : PStore => upd_ps ps x t v)  as f.

assert (NTH: nth (pred (length (skipn i ol))) (map f (skipn i ol)) init_ps
= 
nth (pred (length (skipn i ol))) (map f (skipn i ol)) (f init_ps)).
{ apply nth_indep.
rewrite map_length.
lia. }


rewrite NTH.
rewrite map_nth.

desf.
+

unfold upd_ps, upd.
desf.
+ 
specialize (SAME_END _ _ P2 y).
unfold upd_ps, upd.
desf.


Qed.

Lemma SRAL_step_preserves_wf st st' o (WF: wf_state st) 
  (STEP: SRAL_step st st' o): wf_state st'.
Proof.
inversion STEP; subst; clear STEP; eauto.
eapply SRAL_write_step_helper_preserves_wf; eauto.
eapply SRAL_write_step_helper_preserves_wf; eauto.
Qed.

(******************************************************************************)
(** * Initial state  *)
(******************************************************************************)

(* Definition SRALinit (st : State) := 
  st = fun p => [[fun x => RT T0 0 true]].
Lemma SRALinit_wf (st : State) (INIT : SRALinit st) : wf_state st.
Proof.
unfold SRALinit in *; desf.
red; splits.
- desf. 
- red; ins; splits; ins; desf; ins; desf.
- ins. exfalso. destruct n2; ins; desf.
  lia. destruct n2; ins; desf.
- eexists _,_; splits; ins; desf; ins; desf.
Qed.
 *)

Definition SRALinit (st : State) := forall t,
  << NON_EMPTY_POT : st t <> [] >> /\
  << NON_EMPTY_OL  : forall L (IN: In L (st t)), L <> [] >> /\
  << INIT : forall L (IN: In L (st t)), 
          Forall (eq (fun x => RT T0 0 true)) L >>.

Lemma SRALinit_wf (st : State) (INIT : SRALinit st) : wf_state st.
Proof.
unfold SRALinit in *; desf.
red; splits.
- ins; apply INIT.
- ins; eapply INIT; edone.
- ins.
apply nth_error_In in IN2.
eapply Forall_in in IN2.
2: eby eapply INIT.
desf.
- exists (fun x => 0),(fun x => T0). 
splits; ins; desf; ins; desf.
specialize (INIT t); desc.
specialize (INIT _ IN).
eapply Forall_nth in INIT.
by rewrite <- INIT.
destruct L; try (ins; lia).
exfalso; eapply NON_EMPTY_OL; eauto.
Qed.

End SRALn.

